@@ -12,21 +12,6 @@ static void mcp_funcgen_cleanup(lua_State *L, mcp_funcgen_t *fgen);
 static void mcp_resume_rctx_from_cb(mcp_rcontext_t *rctx);
 static void proxy_return_rqu_cb(io_pending_t *pending);
 
-static inline void _mcp_queue_hack(conn *c) {
-    if (c) {
-        // HACK
-        // see notes above proxy_run_rcontext.
-        // in case the above resume calls queued new work, we have to submit
-        // it to the backend handling system here.
-        for (io_queue_t *q = c->io_queues; q->type != IO_QUEUE_NONE; q++) {
-            if (q->stack_ctx != NULL) {
-                io_queue_cb_t *qcb = thread_io_queue_get(c->thread, q->type);
-                qcb->submit_cb(q);
-            }
-        }
-    }
-}
-
 // If we're GC'ed but not closed, it means it was created but never
 // attached to a function, so ensure everything is closed properly.
 int mcplib_funcgen_gc(lua_State *L) {
@@ -86,7 +71,7 @@ static void mcp_rcontext_cleanup(lua_State *L, mcp_funcgen_t *fgen, mcp_rcontext
     // cleanup of request queue entries. recurse funcgen cleanup.
     for (int x = 0; x < fgen->max_queues; x++) {
         struct mcp_rqueue_s *rqu = &rctx->qslots[x];
-        if (rqu->obj_type == RQUEUE_TYPE_POOL) {
+        if (rqu->obj_type == RQUEUE_TYPE_POOL || rqu->obj_type == RQUEUE_TYPE_INT) {
             // nothing to do.
         } else if (rqu->obj_type == RQUEUE_TYPE_FGEN) {
             // don't need to recurse, just free the subrctx.
@@ -109,6 +94,19 @@ static void mcp_rcontext_cleanup(lua_State *L, mcp_funcgen_t *fgen, mcp_rcontext
         }
     }
 
+    // look for rctx-local objects.
+    if (rctx->uobj_count) {
+        int lim = fgen->max_queues + rctx->uobj_count;
+        for (int x = fgen->max_queues; x < lim; x++) {
+            struct mcp_rqueue_s *rqu = &rctx->qslots[x];
+            // Don't need to look at the type:
+            // - slot has to be freed (thus cleaned up) before getting here
+            // - any uobj is ref'ed into obj_ref
+            luaL_unref(L, LUA_REGISTRYINDEX, rqu->obj_ref);
+            rqu->obj_ref = 0;
+        }
+    }
+
     // nuke alarm if set.
     // should only be paranoia here, but just in case.
     if (event_pending(&rctx->timeout_event, EV_TIMEOUT, NULL)) {
@@ -135,9 +133,11 @@ static void mcp_rcontext_cleanup(lua_State *L, mcp_funcgen_t *fgen, mcp_rcontext
 // require fewer test rounds for unit tests.
 #define FGEN_FREE_PRESSURE_MAX 100
 #define FGEN_FREE_PRESSURE_DROP 10
+#define FGEN_FREE_WAIT 0
 #else
-#define FGEN_FREE_PRESSURE_MAX 4000
-#define FGEN_FREE_PRESSURE_DROP 100
+#define FGEN_FREE_PRESSURE_MAX 5000
+#define FGEN_FREE_PRESSURE_DROP 200
+#define FGEN_FREE_WAIT 60 // seconds.
 #endif
 static void _mcplib_funcgen_cache(mcp_funcgen_t *fgen, mcp_rcontext_t *rctx) {
     bool do_cache = true;
@@ -148,23 +148,29 @@ static void _mcplib_funcgen_cache(mcp_funcgen_t *fgen, mcp_rcontext_t *rctx) {
     // - If free rctx are less than half of total, reduce pressure.
     // - If pressure is too high, immediately free the rctx, then drop the
     // pressure slightly.
+    // - If pressure is too high, and has been for more than FGEN_FREE_WAIT
+    // seconds, immediately free the rctx, then drop the pressure slightly.
     //
     // This should allow bursty traffic to avoid spinning on alloc/frees,
-    // while one-time bursts will slowly free slots back down to a minimum of
-    // 1.
+    // while one-time bursts will slowly free slots back down to a min of 1.
     if (fgen->free > fgen->total/2 - 1) {
         if (fgen->free_pressure++ > FGEN_FREE_PRESSURE_MAX) {
+            struct timespec now;
+            clock_gettime(CLOCK_REALTIME, &now);
+            if (fgen->free_waiter.tv_sec == 0) {
+                fgen->free_waiter.tv_sec = now.tv_sec + FGEN_FREE_WAIT;
+            }
+
+            if (now.tv_sec >= fgen->free_waiter.tv_sec) {
+                do_cache = false;
+            }
+            // check again in a little while.
             fgen->free_pressure -= FGEN_FREE_PRESSURE_DROP;
-            // do not cache the rctx
-            assert(fgen->self_ref);
-            lua_State *L = fgen->thread->L;
-            lua_rawgeti(L, LUA_REGISTRYINDEX, fgen->self_ref);
-            mcp_rcontext_cleanup(L, fgen, rctx, lua_absindex(L, -1));
-            lua_pop(L, 1); // drop fgen
-            do_cache = false;
         }
     } else {
         fgen->free_pressure >>= 1;
+        // must be too-free for a full wait period before releasing.
+        fgen->free_waiter.tv_sec = 0;
     }
 
     if (do_cache) {
@@ -178,6 +184,13 @@ static void _mcplib_funcgen_cache(mcp_funcgen_t *fgen, mcp_rcontext_t *rctx) {
         }
         fgen->list[fgen->free] = rctx;
         fgen->free++;
+    } else {
+        // do not cache the rctx
+        assert(fgen->self_ref);
+        lua_State *L = fgen->thread->L;
+        lua_rawgeti(L, LUA_REGISTRYINDEX, fgen->self_ref);
+        mcp_rcontext_cleanup(L, fgen, rctx, lua_absindex(L, -1));
+        lua_pop(L, 1); // drop fgen
     }
 
     // we're closed and every outstanding request slot has been
@@ -192,7 +205,8 @@ static int _mcplib_funcgen_gencall(lua_State *L) {
     mcp_funcgen_t *fgen = luaL_checkudata(L, -2, "mcp.funcgen");
     int fgen_idx = lua_absindex(L, -2);
     // create the ctx object.
-    size_t rctx_len = sizeof(mcp_rcontext_t) + sizeof(struct mcp_rqueue_s) * fgen->max_queues;
+    int total_queues = fgen->max_queues + fgen->uobj_queues;
+    size_t rctx_len = sizeof(mcp_rcontext_t) + sizeof(struct mcp_rqueue_s) * total_queues;
     mcp_rcontext_t *rc = lua_newuserdatauv(L, rctx_len, 0);
     memset(rc, 0, rctx_len);
 
@@ -207,7 +221,7 @@ static int _mcplib_funcgen_gencall(lua_State *L) {
         struct mcp_rqueue_s *frqu = &fgen->queue_list[x];
         struct mcp_rqueue_s *rqu = &rc->qslots[x];
         rqu->obj_type = frqu->obj_type;
-        if (frqu->obj_type == RQUEUE_TYPE_POOL) {
+        if (frqu->obj_type == RQUEUE_TYPE_POOL || frqu->obj_type == RQUEUE_TYPE_INT) {
             rqu->obj_ref = 0;
             rqu->obj = frqu->obj;
             mcp_resp_t *r = mcp_prep_bare_resobj(L, fgen->thread);
@@ -299,26 +313,23 @@ static int _mcplib_funcgen_gencall(lua_State *L) {
 static void _mcp_funcgen_return_rctx(mcp_rcontext_t *rctx) {
     mcp_funcgen_t *fgen = rctx->fgen;
     assert(rctx->pending_reqs == 0);
-    int res = lua_resetthread(rctx->Lc);
+    int res = lua_status(rctx->Lc);
     if (res != LUA_OK) {
-        // TODO: I was under the impression it was possible to reuse a
-        // coroutine from an error state, but it seems like this only works if
-        // the routine landed in LUA_YIELD or LUA_OK
-        // Leaving a note here to triple check this or if my memory was wrong.
-        // Instead for now we throw away the coroutine if it was involved in
-        // an error. Realistically this shouldn't be normal so it might not
-        // matter anyway.
+        // Can't reuse the thread if we ended in an error.
+        // Reset and close out the old thread.
+        lua_resetthread(rctx->Lc);
         lua_State *L = fgen->thread->L;
         luaL_unref(L, LUA_REGISTRYINDEX, rctx->coroutine_ref);
+        // Make a new thread.
         rctx->Lc = lua_newthread(L);
         assert(rctx->Lc);
         rctx->coroutine_ref = luaL_ref(L, LUA_REGISTRYINDEX);
     } else {
+        // Thread is okay, clear stack and continue.
         lua_settop(rctx->Lc, 0);
     }
     rctx->wait_mode = QWAIT_IDLE;
     rctx->resp = NULL;
-    rctx->first_queue = false; // HACK
     if (rctx->request) {
         mcp_request_cleanup(fgen->thread, rctx->request);
     }
@@ -353,6 +364,24 @@ static void _mcp_funcgen_return_rctx(mcp_rcontext_t *rctx) {
             _mcp_funcgen_return_rctx(rqu->obj);
         }
     }
+
+    // look for rctx-local objects.
+    if (rctx->uobj_count) {
+        int lim = fgen->max_queues + rctx->uobj_count;
+        for (int x = fgen->max_queues; x < lim; x++) {
+            struct mcp_rqueue_s *rqu = &rctx->qslots[x];
+            if (rqu->obj_type == RQUEUE_TYPE_UOBJ_REQ) {
+                mcp_request_t *rq = rqu->obj;
+                mcp_request_cleanup(fgen->thread, rq);
+            } else if (rqu->obj_type == RQUEUE_TYPE_UOBJ_RES) {
+                mcp_resp_t *rs = rqu->obj;
+                mcp_response_cleanup(fgen->thread, rs);
+            } else {
+                // no known type. only crash the debug binary.
+                assert(1 == 0);
+            }
+        }
+    }
 }
 
 // TODO: check rctx->awaiting before returning?
@@ -385,6 +414,7 @@ mcp_rcontext_t *mcp_funcgen_get_rctx(lua_State *L, int fgen_ref, mcp_funcgen_t *
     if (fgen->free == 0) {
         // reset free pressure so we try to keep the rctx cached
         fgen->free_pressure = 0;
+        fgen->free_waiter.tv_sec = 0;
         // TODO (perf): pre-create this c closure somewhere hidden.
         lua_pushcclosure(L, _mcplib_funcgen_gencall, 0);
         // pull in the funcgen object
@@ -431,7 +461,7 @@ mcp_rcontext_t *mcp_funcgen_start(lua_State *L, mcp_funcgen_t *fgen, mcp_parser_
     // Note that we can _technically_ fail with an OOM here, but we've not set
     // up lua in a way that OOM's are possible.
     if (rctx->request_ref == 0) {
-        mcp_request_t *rq = lua_newuserdatauv(L, sizeof(mcp_request_t) + MCP_REQUEST_MAXLEN + KEY_MAX_LENGTH, 0);
+        mcp_request_t *rq = lua_newuserdatauv(L, sizeof(mcp_request_t) + MCP_REQUEST_MAXLEN, 0);
         memset(rq, 0, sizeof(mcp_request_t));
         luaL_getmetatable(L, "mcp.request");
         lua_setmetatable(L, -2);
@@ -498,7 +528,7 @@ static void mcp_funcgen_cleanup(lua_State *L, mcp_funcgen_t *fgen) {
     if (fgen->queue_list) {
         for (int x = 0; x < fgen->max_queues; x++) {
             struct mcp_rqueue_s *rqu = &fgen->queue_list[x];
-            if (rqu->obj_type == RQUEUE_TYPE_POOL) {
+            if (rqu->obj_type == RQUEUE_TYPE_POOL || rqu->obj_type == RQUEUE_TYPE_INT) {
                 // just the obj_ref
                 luaL_unref(L, LUA_REGISTRYINDEX, rqu->obj_ref);
             } else if (rqu->obj_type == RQUEUE_TYPE_FGEN) {
@@ -609,6 +639,7 @@ int mcplib_funcgen_new_handle(lua_State *L) {
     mcp_funcgen_t *fgen = lua_touserdata(L, 1);
     mcp_pool_proxy_t *pp = NULL;
     mcp_funcgen_t *fg = NULL;
+    void *test = NULL;
 
     if (fgen->ready) {
         proxy_lua_error(L, "cannot modify function generator after calling ready");
@@ -617,6 +648,8 @@ int mcplib_funcgen_new_handle(lua_State *L) {
 
     if ((pp = luaL_testudata(L, 2, "mcp.pool_proxy")) != NULL) {
         // good.
+    } else if ((test = luaL_testudata(L, 2, "mcp.internal_be")) != NULL) {
+        // also good.
     } else if ((fg = luaL_testudata(L, 2, "mcp.funcgen")) != NULL) {
         if (fg->is_router) {
             proxy_lua_error(L, "cannot assign a router to a handle in new_handle");
@@ -650,6 +683,11 @@ int mcplib_funcgen_new_handle(lua_State *L) {
         rqu->obj_ref = luaL_ref(L, LUA_REGISTRYINDEX);
         rqu->obj_type = RQUEUE_TYPE_POOL;
         rqu->obj = pp;
+    } else if (test) {
+        // pops test from the stack
+        rqu->obj_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+        rqu->obj_type = RQUEUE_TYPE_INT;
+        rqu->obj = test;
     } else {
         // pops the fgen from the stack.
         mcp_funcgen_reference(L);
@@ -691,6 +729,16 @@ int mcplib_funcgen_ready(lua_State *L) {
         lua_pop(L, 1);
     }
 
+    if (lua_getfield(L, 2, "u") == LUA_TNUMBER) {
+        int uobj_queues = luaL_checkinteger(L, -1);
+        if (uobj_queues < 1 || uobj_queues > RQUEUE_UOBJ_MAX) {
+            proxy_lua_ferror(L, "user obj ('u') in fgen:ready must be between 1 and %d", RQUEUE_UOBJ_MAX);
+            return 0;
+        }
+        fgen->uobj_queues = uobj_queues;
+    }
+    lua_pop(L, 1);
+
     // now we test the generator function and create the first slot.
     lua_pushvalue(L, 1); // copy the funcgen to pass into gencall
     lua_rawgeti(L, LUA_REGISTRYINDEX, fgen->generator_ref); // for gencall
@@ -886,6 +934,8 @@ static void mcp_resume_rctx_from_cb(mcp_rcontext_t *rctx) {
     if (rctx->parent) {
         struct mcp_rqueue_s *rqu = &rctx->parent->qslots[rctx->parent_handle];
         if (res == LUA_OK) {
+            mcp_rcontext_t *parent = rctx->parent;
+            int handle = rctx->parent_handle;
             int type = lua_type(rctx->Lc, 1);
             mcp_resp_t *r = NULL;
             if (type == LUA_TUSERDATA && (r = luaL_testudata(rctx->Lc, 1, "mcp.response")) != NULL) {
@@ -904,40 +954,22 @@ static void mcp_resume_rctx_from_cb(mcp_rcontext_t *rctx) {
             } else {
                 // generate a generic object with an error.
                 _mcp_resume_rctx_process_error(rctx, rqu);
+                mcp_funcgen_return_rctx(rctx);
+                return;
             }
-            if (rctx->parent->wait_count) {
-                mcp_process_rctx_wait(rctx->parent, rctx->parent_handle);
-            }
+
+            // return ourself before telling the parent to wait.
             mcp_funcgen_return_rctx(rctx);
+            if (parent->wait_count) {
+                mcp_process_rctx_wait(parent, handle);
+            }
         } else if (res == LUA_YIELD) {
             // normal.
-            _mcp_queue_hack(rctx->c);
         } else {
             lua_pop(rctx->Lc, 1); // drop the error message.
             _mcp_resume_rctx_process_error(rctx, rqu);
             mcp_funcgen_return_rctx(rctx);
         }
-    } else {
-        // Parent rctx has returned either a response or error to its top
-        // level resp object and is now complete.
-        // HACK
-        // see notes in proxy_run_rcontext()
-        // NOTE: this function is called from rqu_cb(), which pushes the
-        // submission loop. This code below can call drive_machine(), which
-        // calls submission loop if stuff is queued.
-        // Would remove redundancy if we can signal up to rqu_cb() and either
-        // q->count-- or do the inline submission at that level.
-        if (res != LUA_YIELD) {
-            mcp_funcgen_return_rctx(rctx);
-            io_queue_t *q = conn_io_queue_get(rctx->c, IO_QUEUE_PROXY);
-            q->count--;
-            if (q->count == 0) {
-                // call re-add directly since we're already in the worker thread.
-                conn_worker_readd(rctx->c);
-            }
-        } else if (res == LUA_YIELD) {
-            _mcp_queue_hack(rctx->c);
-        }
     }
 }
 
@@ -1013,7 +1045,29 @@ void mcp_process_rctx_wait(mcp_rcontext_t *rctx, int handle) {
     }
 
     assert(rctx->pending_reqs != 0);
-    if ((status & RQUEUE_R_RESUME) || rctx->wait_done == rctx->wait_count || rctx->pending_reqs == 1) {
+    bool should_resume = (status & RQUEUE_R_RESUME);
+    if (rctx->wait_done == rctx->wait_count || rctx->pending_reqs == 1) {
+        should_resume = true;
+    } else if (status & RQUEUE_R_ERROR) {
+        // An error condition could happen without the pending req
+        // decrementing. Since this should be a rare state we do a slow check
+        // on if the wait result is still possible.
+        int possible = 0;
+        // move an errored RQU to a WAITED state so we don't block the wait on
+        // it during a future error.
+        rqu->state = RQUEUE_WAITED;
+        for (int x = 0; x < rctx->fgen->max_queues; x++) {
+            struct mcp_rqueue_s *rqu = &rctx->qslots[x];
+            if (rqu->state != RQUEUE_IDLE && rqu->state != RQUEUE_WAITED) {
+                possible++;
+            }
+        }
+        if (possible == 0) {
+            should_resume = true;
+        }
+    }
+
+    if (should_resume) {
         // ran out of stuff to wait for. time to resume.
         // TODO: can we do the settop at the yield? nothing we need to
         // keep in the stack in this mode.
@@ -1050,6 +1104,12 @@ int mcp_process_rqueue_return(mcp_rcontext_t *rctx, int handle, mcp_resp_t *res)
         } else {
             flag = RQUEUE_R_OK;
         }
+    } else {
+        flag = RQUEUE_R_ERROR;
+    }
+
+    if (res->be && res->be->use_logging) {
+        mcplib_rqu_log(rqu->rq, res, flag, rctx->conn_fd);
     }
 
     if (rqu->cb_ref) {
@@ -1085,6 +1145,7 @@ int mcp_process_rqueue_return(mcp_rcontext_t *rctx, int handle, mcp_resp_t *res)
                                  // we settop _before_ calling cb's and
                                  // _before_ setting up for a coro resume.
     }
+
     rqu->flags |= flag;
     return rqu->flags;
 }
@@ -1118,15 +1179,50 @@ void mcp_run_rcontext_handle(mcp_rcontext_t *rctx, int handle) {
         if (rqu->obj_type == RQUEUE_TYPE_POOL) {
             mcp_request_t *rq = rqu->rq;
             mcp_backend_t *be = mcplib_pool_proxy_call_helper(rqu->obj, MCP_PARSER_KEY(rq->pr), rq->pr.klen);
-            // FIXME: queue requires conn because we're stacking objects
-            // into the connection for later submission, which means we
-            // absolutely cannot queue things once *c becomes invalid.
-            // need to assert/block this from happening.
+
             mcp_set_resobj(rqu->res_obj, rq, be, rctx->fgen->thread);
             io_pending_proxy_t *p = mcp_queue_rctx_io(rctx, rq, be, rqu->res_obj);
             p->return_cb = proxy_return_rqu_cb;
             p->queue_handle = handle;
             rctx->pending_reqs++;
+        } else if (rqu->obj_type == RQUEUE_TYPE_INT) {
+            mcp_request_t *rq = rqu->rq;
+            mc_resp *resp = mcp_rcontext_internal(rctx, rq, rqu->res_obj);
+            if (resp == NULL) {
+                // NOTE: This can be OOM (no resp alloc)
+                // or bad parse (no such command)
+                // we _could_ set an ERRMSG here.
+                mcp_resp_t *r = rqu->res_obj;
+                r->status = MCMC_ERR;
+                r->resp.code = MCMC_CODE_SERVER_ERROR;
+                io_pending_proxy_t *p = mcp_queue_rctx_io(rctx, NULL, NULL, rqu->res_obj);
+                p->return_cb = proxy_return_rqu_cb;
+                p->queue_handle = handle;
+                p->background = true;
+                rctx->pending_reqs++;
+            } else if (resp->io_pending) {
+                resp->io_pending->return_cb = proxy_return_rqu_cb;
+                // Add io object to extstore submission queue.
+                io_queue_t *q = thread_io_queue_get(rctx->fgen->thread, IO_QUEUE_EXTSTORE);
+                io_pending_proxy_t *io = (io_pending_proxy_t *)resp->io_pending;
+                io->queue_handle = handle;
+                io->client_resp = rqu->res_obj;
+
+                STAILQ_INSERT_TAIL(&q->stack, (io_pending_t *)io, iop_next);
+
+                io->rctx = rctx;
+                io->c = rctx->c;
+                io->ascii_multiget = rq->ascii_multiget;
+                // mark the buffer into the mcp_resp for freeing later.
+                rqu->res_obj->buf = io->eio.buf;
+                rctx->pending_reqs++;
+            } else {
+                io_pending_proxy_t *p = mcp_queue_rctx_io(rctx, NULL, NULL, rqu->res_obj);
+                p->return_cb = proxy_return_rqu_cb;
+                p->queue_handle = handle;
+                p->background = true;
+                rctx->pending_reqs++;
+            }
         } else if (rqu->obj_type == RQUEUE_TYPE_FGEN) {
             // TODO: NULL the ->c post-return?
             mcp_rcontext_t *subrctx = rqu->obj;
@@ -1303,32 +1399,6 @@ int mcplib_rcontext_wait_handle(lua_State *L) {
     return lua_yield(L, 1);
 }
 
-// TODO: This is disabled due to issues with the IO subsystem. Fixing this
-// requires retiring of API1 to allow refactoring or some extra roundtrip
-// work.
-// - if rctx:sleep() is called, the coroutine is suspended.
-// - once resumed after the timeout, we may later suspend again and make
-// backend requests
-// - once the coroutine is completed, we need to check if the owning client
-// conn is ready to be resumed
-// - BUG: we can only get into the "conn is in IO queue wait" state if a
-// sub-IO was created and submitted somewhere.
-// - This means either rctx:sleep() needs to be implemented by submitting a
-// dummy IO req (as other code does)
-// - OR we need to refactor the IO system so the dummies aren't required
-// anymore.
-//
-// If a dummy is used, we would have to implement this as:
-// - immediately submit a dummy IO if sleep is called.
-// - this allows the IO system to move the connection into the right state
-// - will immediately circle back then set an alarm for the sleep timeout
-// - once the sleep resumes, run code as normal. resumption should work
-// properly since we've entered the correct state originally.
-//
-// This adds a lot of CPU overhead to sleep, which should be fine given the
-// nature of the function, but also adds a lot of code and increases the
-// chances of bugs. So I'm leaving it out until this can be implemented more
-// simply.
 int mcplib_rcontext_sleep(lua_State *L) {
     mcp_rcontext_t *rctx = lua_touserdata(L, 1);
     if (rctx->wait_mode != QWAIT_IDLE) {
@@ -1403,6 +1473,98 @@ int mcplib_rcontext_result(lua_State *L) {
     return 2;
 }
 
+// arg must be an array table.
+// returns res, GOOD|OK|ANY
+// tries to find a result in that order.
+int mcplib_rcontext_best_result(lua_State *L) {
+    mcp_rcontext_t *rctx = lua_touserdata(L, 1);
+
+    if (lua_istable(L, 2)) {
+        int final_handle = -1;
+        int final_flag = -1;
+        unsigned int len = lua_rawlen(L, 2);
+        for (int x = 0; x < len; x++) {
+            lua_rawgeti(L, 2, x+1);
+            int handle = lua_tointeger(L, 3);
+            lua_pop(L, 1);
+
+            if (handle < 0 || handle >= rctx->fgen->max_queues) {
+                proxy_lua_error(L, "invalid queue handle passed to best_result");
+            }
+
+            struct mcp_rqueue_s *rqu = &rctx->qslots[handle];
+            if (!rqu->flags) {
+                continue; // error or unprocessed.
+            } else if (rqu->flags & RQUEUE_R_GOOD) {
+                final_handle = handle;
+                break;
+            } else if (rqu->flags & RQUEUE_R_OK) {
+                final_handle = handle;
+                final_flag = RQUEUE_R_OK;
+            } else if (final_flag != RQUEUE_R_OK) {
+                // only use an error if we don't already have an OK
+                final_handle = handle;
+            }
+        }
+
+        if (final_handle != -1) {
+            struct mcp_rqueue_s *rqu = &rctx->qslots[final_handle];
+            lua_rawgeti(L, LUA_REGISTRYINDEX, rqu->res_ref);
+            lua_pushinteger(L, rqu->flags & (RQUEUE_R_ANY|RQUEUE_R_OK|RQUEUE_R_GOOD));
+        } else {
+            lua_pushnil(L);
+            lua_pushnil(L);
+        }
+    } else {
+        proxy_lua_error(L, "must pass a table to :best_result");
+    }
+    return 2;
+}
+
+// arg must be an array table.
+// returns res, ANY|OK|GOOD
+// tries to find a result in that order.
+// TODO: test with incomplete requests? not sure what the data looks like.
+int mcplib_rcontext_worst_result(lua_State *L) {
+    mcp_rcontext_t *rctx = lua_touserdata(L, 1);
+
+    if (lua_istable(L, 2)) {
+        int final_handle = -1;
+        int final_flags = 0;
+        unsigned int len = lua_rawlen(L, 2);
+        for (int x = 0; x < len; x++) {
+            lua_rawgeti(L, 2, x+1);
+            int handle = lua_tointeger(L, 3);
+            lua_pop(L, 1);
+
+            if (handle < 0 || handle >= rctx->fgen->max_queues) {
+                proxy_lua_error(L, "invalid queue handle passed to worst_result");
+            }
+
+            struct mcp_rqueue_s *rqu = &rctx->qslots[handle];
+            if (!rqu->flags) {
+                continue;
+            } else if (final_flags <= rqu->flags) {
+                // flag values increase by how bad they are.
+                final_handle = handle;
+                final_flags = rqu->flags;
+            }
+        }
+
+        if (final_handle != -1) {
+            struct mcp_rqueue_s *rqu = &rctx->qslots[final_handle];
+            lua_rawgeti(L, LUA_REGISTRYINDEX, rqu->res_ref);
+            lua_pushinteger(L, rqu->flags & (RQUEUE_R_ANY|RQUEUE_R_OK|RQUEUE_R_GOOD));
+        } else {
+            lua_pushnil(L);
+            lua_pushnil(L);
+        }
+    } else {
+        proxy_lua_error(L, "must pass a table to :worst_result");
+    }
+    return 2;
+}
+
 int mcplib_rcontext_cfd(lua_State *L) {
     mcp_rcontext_t *rctx = lua_touserdata(L, 1);
     lua_pushinteger(L, rctx->conn_fd);
@@ -1431,22 +1593,47 @@ int mcplib_rcontext_tls_peer_cn(lua_State *L) {
     return 1;
 }
 
-// TODO: stub function.
-// need to attach request function to rcontext, by using another function that
-// doesn't fill out the request info
-int mcplib_rcontext_request_new(lua_State *L) {
-    mcp_parser_t pr = {0};
+// call with uobj on top of stack
+static void _mcplib_rcontext_ref_uobj(lua_State *L, mcp_rcontext_t *rctx, void *obj, int otype) {
+    lua_pushvalue(L, -1); // dupe rq for the rqueue slot
+    struct mcp_rqueue_s *rqu = &rctx->qslots[rctx->fgen->max_queues + rctx->uobj_count];
+    rctx->uobj_count++;
+    // hold the request reference into the rctx for memory management.
+    rqu->obj_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+    rqu->obj_type = otype;
+    rqu->obj = obj;
+}
 
-    mcp_new_request(L, &pr, " ", 1);
+// Creates request object that's tracked by request context so we can call
+// cleanup routines post-run.
+int mcplib_rcontext_request_new(lua_State *L) {
+    mcp_rcontext_t *rctx = lua_touserdata(L, 1);
+    if (rctx->uobj_count == rctx->fgen->uobj_queues) {
+        proxy_lua_error(L, "rctx request new: object count limit reached");
+        return 0;
+    }
+
+    // create new request object
+    mcp_parser_t pr = {0};
+    mcp_request_t *rq = mcp_new_request(L, &pr, " ", 1);
+
+    _mcplib_rcontext_ref_uobj(L, rctx, rq, RQUEUE_TYPE_UOBJ_REQ);
     return 1;
 }
 
-// TODO: stub function, see request_new above.
 int mcplib_rcontext_response_new(lua_State *L) {
+    mcp_rcontext_t *rctx = lua_touserdata(L, 1);
+    if (rctx->uobj_count == rctx->fgen->uobj_queues) {
+        proxy_lua_error(L, "rctx request new: object count limit reached");
+        return 0;
+    }
+
     mcp_resp_t *r = lua_newuserdatauv(L, sizeof(mcp_resp_t), 0);
     memset(r, 0, sizeof(mcp_resp_t));
     luaL_getmetatable(L, "mcp.response");
     lua_setmetatable(L, -2);
+
+    _mcplib_rcontext_ref_uobj(L, rctx, r, RQUEUE_TYPE_UOBJ_RES);
     return 1;
 }
 
@@ -1506,7 +1693,7 @@ static inline const char *_mcp_router_longsep(const char *key, const int klen, c
     // find the last possible position
     const char *last = key + (klen - nlen);
 
-    while (end != last) {
+    while (end <= last) {
         if (*end == needle[0] && memcmp(end, needle, nlen) == 0) {
             lookup = key;
             *len = end - key;
